import 'package:flutter/services.dart';
import 'package:nend_plugin/nend_plugin.dart';
import 'dart:async';
import 'package:flutter/material.dart';

class NativeAdLoader {
    static const String tag = 'NendAdNativeClient';
    static Set<String> _registeredAdunit = Set<String>();

    final int spotId;
    final String apiKey;
    String get _mappingId => this.hashCode.toString();
    Map simpleMappingArgument;
    final NativeAdLoaderListener listener;

    NativeAdLoader(this.spotId, this.apiKey, this.listener) {
        if (_registeredAdunit.contains(this.spotId.toString())) {
            var message = "This ad unit already created. Please release old object before create new one.";
            throw new StateError(message);
        }
        _registeredAdunit.add(this.spotId.toString());
        NendPlugin.invoke(_mappingId, NativeAdLoader.tag, NendPlugin.method_name_init, argument: {
            NendPlugin.key_mapping_id: _mappingId,
            NendPlugin.key_ad_unit: {NendPlugin.key_spot_id: spotId, NendPlugin.key_api_key: apiKey}
        });
        NendPlugin.addListener(_mappingId, this);
    }

    Future<void> loadAd({int intervalMillis}) async {
        if (intervalMillis == null || intervalMillis < 30000) {
            NendPlugin.invoke(_mappingId, NativeAdLoader.tag, NendPlugin.method_name_load_ad);
        } else {
            autoReloadIntervalMillis = intervalMillis;
        }
    }

    NativeAd _parseNativeAd(Map<String, dynamic> map) {
        var adMap = map['nativeAd'];
        NativeAd nativeAd = NativeAd._internal(
            adMap['titleText'],
            adMap['contentText'],
            adMap['promotionName'],
            adMap['promotionUrl'],
            adMap['adImageUrl'],
            adMap['logoImageUrl'],
            adMap['actionButtonText'],
            adMap['hashCode']
        );
        return nativeAd;
    }

    int _autoReloadIntervalMillis;
    get autoReloadIntervalMillis => _autoReloadIntervalMillis;
    set autoReloadIntervalMillis(int millisec) {
        _autoReloadIntervalMillis = millisec;
        if (millisec < 30000) {
            NendPlugin.invoke(_mappingId, NativeAdLoader.tag, '.disableAutoReload');
        } else {
            _enableAutoReload();
        }
    }

    Future<void> _enableAutoReload() async {
        NendPlugin.invoke(_mappingId, tag, '.enableAutoReload', argument: {'intervalMillis': _autoReloadIntervalMillis});
    }

    void releaseLoader() async {
        _registeredAdunit.remove(this.spotId.toString());
        await NendPlugin.invoke(_mappingId, NativeAdLoader.tag, NendPlugin.method_name_dispose);
    }

    dynamic eventFrom(String type) {
        type = type.split(".").last;
        type = 'NativeAdLoaderEvent.$type';
        return NativeAdLoaderEvent.values
            .firstWhere((event) => event.toString() == type, orElse: () => null);
    }

    void notify(MethodCall call) {
        NativeAdLoaderEvent event = eventFrom(call.method);
        Map<String, dynamic> map = call.arguments;
        if (event == NativeAdLoaderEvent.onFailedToLoad) {
            int code = int.parse(map[NendPlugin.key_error_code].toString());
            listener(event, errorCode: code);
        } else {
            listener(event, ad: _parseNativeAd(call.arguments));
        }
    }

}

enum NativeAdLoaderEvent {
    onLoaded,
    onFailedToLoad,
}

typedef void NativeAdLoaderListener(NativeAdLoaderEvent event, {int errorCode, NativeAd ad});

class NativeAd {
    static const String tag = 'NendNativeAdConnector';

    final String _mappingId;
    final String titleText;
    final String contentText;
    final String promotionName;
    final String promotionUrl;
    final String adImageUrl;
    final String logoImageUrl;
    final String actionButtonText;

    bool _isActivated = false;

    factory NativeAd() {
        var message = "Not support this operation.";
        throw new StateError(message);
    }

    NativeAd._internal(
        this.titleText,
        this.contentText,
        this.promotionName,
        this.promotionUrl,
        this.adImageUrl,
        this.logoImageUrl,
        this.actionButtonText,
        //最後にmappingId
        this._mappingId
    );

    NativeAdListener _listener;
    NativeAdListener get listener => _listener;
    set listener(NativeAdListener value) {
        _listener = value;
        NendPlugin.addListener(tag, this);
    }

    static String _getAdvertisingExplicitlyText(AdvertisingExplicitly type) {
        switch (type) {
            case AdvertisingExplicitly.PR:
                return "PR";
            case AdvertisingExplicitly.Sponsored:
                return "Sponsored";
            case AdvertisingExplicitly.AD:
                return "広告";
            case AdvertisingExplicitly.Promotion:
                return "プロモーション";
            default:
                throw new StateError("Unknown AdvertisingExplicitly type...");
        }
    }

    Widget getAdvertisingExplicitlyWidget(AdvertisingExplicitly type, {TextStyle style}) {
        Widget text = (style == null
            ? Text(NativeAd._getAdvertisingExplicitlyText(type))
            : Text(NativeAd._getAdvertisingExplicitlyText(type), style: style)
        );
        return GestureDetector(
            onTap: () {
                _performInformationClick();
            },
            child: Container(
                child: text,
            ),
        );

    }

    void _performAdClick() {
        if (_isActivated) {
            NendPlugin.invoke(_mappingId, NativeAd.tag, '.performAdClick');
            if (_listener != null) {
                listener(NativeAdEvent.onClickAd);
            }
        }
    }

    void _performInformationClick() {
        NendPlugin.invoke(_mappingId, NativeAd.tag, '.performInformationClick');
        if (_listener != null) {
            listener(NativeAdEvent.onClickInformation);
        }
    }

    void _activate() {
        if (!_isActivated) {
            NendPlugin.invoke(_mappingId, NativeAd.tag, '.activate');
            _isActivated = true;
            if (_listener != null) {
                listener(NativeAdEvent.onImpression);
            }
        }
    }

    void _deactivate() {
        NendPlugin.invoke(_mappingId, NativeAd.tag, NendPlugin.method_name_dispose);
    }
}

enum NativeAdEvent {
    onImpression,
    onClickAd,
    onClickInformation
}

enum AdvertisingExplicitly {
    PR,
    Sponsored,
    AD,
    Promotion
}

typedef void NativeAdListener(NativeAdEvent event);

class NativeAdBinder extends StatefulWidget {
    final Widget _adContainer;
    final Stream<NativeAd> _streamNativeAd;

    NativeAdBinder(this._streamNativeAd, this._adContainer);

    @override
    _NativeAdBinder createState() => new _NativeAdBinder();
}

class _NativeAdBinder extends State<NativeAdBinder> with WidgetsBindingObserver {
    GlobalKey _globalKey = GlobalKey();
    Timer _viewableChecker;
    NativeAd _nativeAd;

    @override
    Widget build(BuildContext context) {
        return StreamBuilder<NativeAd>(
            stream: widget._streamNativeAd,
            builder: (BuildContext context,  AsyncSnapshot<NativeAd> snapshot) {
                if (snapshot.hasData) {
                    _nativeAd = snapshot.data;
                }
                return Container(
                    key: _globalKey,
                    child: GestureDetector(
                        onTap: () {
                            _nativeAd._performAdClick();
                        },
                        child: widget._adContainer,
                    ),
                );
            }
        );
    }

    @override
    void initState() {
        WidgetsBinding.instance.addPostFrameCallback(_afterLayout);
        WidgetsBinding.instance.addObserver(this);
        super.initState();
    }

    @override
    void dispose() {
        WidgetsBinding.instance.removeObserver(this);
        if (_nativeAd != null) {
            _nativeAd._deactivate();
        }
        _stopViewableChecker();
        super.dispose();
    }

    @override
    void didChangeMetrics() {
        WidgetsBinding.instance.addPostFrameCallback(_afterLayout);
        super.didChangeMetrics();
    }

    _afterLayout(_) {
        if (_nativeAd != null && !_nativeAd._isActivated && _isViewable()) {
            _stopViewableChecker();
            _nativeAd._activate();
        } else if (_viewableChecker == null) {
            _viewableChecker = Timer.periodic(Duration(seconds: 2), (_) {
                _afterLayout(_);
            });
        }
    }

    void _stopViewableChecker() {
        if (_viewableChecker != null) {
            _viewableChecker.cancel();
            _viewableChecker = null;
        }
    }

    bool _isViewable() {
        final RenderBox renderBox = _globalKey.currentContext.findRenderObject();
        final position = renderBox.localToGlobal(Offset.zero);
        //Note: Logging対応したら復活させる -> print('position: $position');
        final binderSize = renderBox.size;
        //Note: Logging対応したら復活させる -> print('binderSize: $binderSize');
        final displaySize = MediaQuery.of(context).size;
        //Note: Logging対応したら復活させる -> print('displaySize: $displaySize');

        Rect rect = Rect.fromLTWH(0, 0, displaySize.width, displaySize.height);
        return rect.contains(position) && _isOver50percent(position, binderSize, displaySize);
    }

    //Note : 親Widgetの描画領域をプラグイン内で取る事ができないので↓の処理は
    //       NavigationBarがあったり、SafeAreaウィジェット内に書かれてたりすると正しく機能できない(;∀;)
    bool _isOver50percent(Offset position, Size binderSize, Size displaySize) {
        bool checkPositionX = (position.dx < 0
            ? binderSize.width / 2 < binderSize.width - position.dx
            : displaySize.width - position.dx > binderSize.width / 2
        );
        bool checkPositionY = (position.dy < 0
            ? binderSize.height / 2 < binderSize.height - position.dy
            : displaySize.height - position.dy > binderSize.height / 2
        );
        //Note: Logging対応したら復活させる -> print('checkPositionX: $checkPositionX');
        //Note: Logging対応したら復活させる -> print('checkPositionY: $checkPositionY');

        return checkPositionX && checkPositionY;
    }
}